from typing import List, Tuple

from gnosis.safe import Safe
from gnosis.eth import EthereumClient

from eulith_web3.erc20 import TokenSymbol, EulithWETH
from eulith_web3.eulith_web3 import EulithWeb3
from eulith_web3.signing import construct_signing_middleware, LocalSigner
from eulith_web3.trezor import TrezorSigner
from eulith_web3.contract_bindings.safe.i_safe import ISafe

EULITH_REFRESH_TOKEN='eyJ0eXAiOiJKV1QiLCJhbGciOiJFUzI1NksifQ.eyJzdWIiOiJsaWJ0ZXN0IiwiZXhwIjoxODQ0Njc0NDA3MzcwOTU1MTYxNSwic291cmNlX2hhc2giOiIqIiwic2NvcGUiOiJBUElSZWZyZXNoIn0.G87Tv9LwLH8SRgjlVKIAPk1pdavVS0xwz3fuB7lxP0Et-pPM7ojQkjC1zlC7zWYUdh9p3GvwX_ROfgSPJsw-Qhw'

RESCUE_RECIPIENT = '0x0CE3977D1707fe3954372877A6AE24075b4A78f4'


def setup_armor(ew3: EulithWeb3, authorized_address, owner_keys) -> (str, str):
    aa, sa = ew3.v0.deploy_new_armor(authorized_address, {
        'from': authorized_address,
        'gas': 2500000
    })

    assert '0x' in aa
    assert '0x' in sa

    owner_addresses = []

    c = 0
    threshold = 2

    for o in owner_keys:
        o_wal = LocalSigner(o)
        owner_addresses.append(o_wal.address)
        if c < threshold:
            status = ew3.v0.submit_enable_module_signature(authorized_address, o_wal)
            assert status
        c += 1

    status = ew3.v0.enable_armor(authorized_address, threshold, owner_addresses, {
        'gas': 500000,
        'from': ew3.wallet_address
    })

    assert status

    safe = ISafe(ew3, ew3.to_checksum_address(sa))

    is_module_enabled = safe.is_module_enabled(ew3.to_checksum_address(aa))
    assert is_module_enabled

    read_owners = safe.get_owners()
    assert len(read_owners) == len(owner_keys)

    read_threshold = safe.get_threshold()
    assert read_threshold == threshold

    return aa, sa


def send_weth_to(ew3: EulithWeb3, recipient):
    weth: EulithWETH = EulithWETH(ew3, ew3.to_checksum_address('0x0d500B1d8E8eF31E21C99d1Db9A6444d3ADf1270'))
    tx = weth.deposit_eth(1, {
        'from': ew3.wallet_address,
        'gas': 200000
    })
    r = ew3.eth.send_transaction(tx)
    print(f'Deposit weth transaction: {r.hex()}')

    tx = weth.transfer_float(recipient, 1, {
        'from': ew3.wallet_address,
        'gas': 200000
    })

    r = ew3.eth.send_transaction(tx)
    print(f'Transfer weth transaction: {r.hex()}')


if __name__ == '__main__':
    raw_owners = ['1e780eaf6fed0ce7aca07418d97df74d5a004e394851b127b8c701b725e8400d', '2f2073f613ff2651f2114e26e8cca6e49a73b719dd728c6915eeb0c20534c32e', '59d060b1d6ae7c494a2c911c200429e02bfb4714d948c684261b696ed7077833', 'a39004547e27a0314969c82662733328fd7ce36b373460687a205a69fbb8b699', 'cf9ac1ec7666006d91173960ede131cc01f1a11eccdfa1ab593385da04878721', 'd81ea7876bd225b70410448225e4121fbeacb2bbc13b49c00396b622e9fb06a1']
    owner_keys = sorted(raw_owners, key=lambda x: int(LocalSigner(x).address, 16))

    print(owner_keys)

    wallet = LocalSigner('0x92db14e403b83dfe3df233f83dfa3a0d7096f21ca9b0d6d6b8d88b2b4ec1564e')
    print(wallet.address)

    ew3 = EulithWeb3("http://localhost:7777/v0", EULITH_REFRESH_TOKEN, signing_middle_ware=construct_signing_middleware(wallet))
    ew3.default_confirmations = 0

    armor, safe = setup_armor(ew3, wallet.address, owner_keys)
    send_weth_to(ew3, ew3.to_checksum_address(safe))

    # WMATIC
    weth: EulithWETH = EulithWETH(ew3, ew3.to_checksum_address('0x0d500B1d8E8eF31E21C99d1Db9A6444d3ADf1270'))

    a, s = ew3.v0.get_armor_and_safe_addresses(wallet.address)

    print(f'Safe address: {s}')

    # assert weth.balance_of_float(ew3.to_checksum_address(RESCUE_RECIPIENT)) == 0

    tx = weth.transfer_float(ew3.to_checksum_address(RESCUE_RECIPIENT), 0.1)
    tx_data = tx.get('data')
    stripped_data = tx_data[2:]

    tx_data = bytearray.fromhex(stripped_data)

    safe = ISafe(ew3, ew3.to_checksum_address(s))

    TO = tx.get('to')
    VALUE = 0
    DATA = tx_data
    OP = 0
    TX_GAS = 0
    BASE_GAS = 0
    GAS_PRICE = 0
    GAS_TOKEN = '0x0000000000000000000000000000000000000000'
    REFUND = '0x0000000000000000000000000000000000000000'
    NONCE = safe.nonce()

    tx_hash = safe.get_transaction_hash(
        TO,
        VALUE,
        DATA,
        OP,
        TX_GAS,
        BASE_GAS,
        GAS_PRICE,
        GAS_TOKEN,
        REFUND,
        NONCE
    )

    signature_bytes = bytearray()

    for o in owner_keys:
        owner = LocalSigner(o)
        owner_ew3 = EulithWeb3("http://localhost:7777/v0", EULITH_REFRESH_TOKEN, signing_middle_ware=construct_signing_middleware(owner))
        owner_ew3.default_confirmations = 0

        hash_sig = owner.sign_msg_hash(tx_hash)
        hash_sig.v = 1
        hash_sig.s = 0
        hash_sig.r = int(owner.address, 16)

        print(f'Signature from {owner.address}: {hash_sig.to_hex()}')
        signature_bytes += hash_sig.to_bytes()

        ew3.eth.send_transaction({
            'to': owner.address,
            'from': ew3.wallet_address,
            'value': 1000000000000000000  # 1e18
        })

        approve_tx = safe.approve_hash(tx_hash, {
            'from': owner.address,
            'gas': 300000
        })
        owner_ew3.eth.send_transaction(approve_tx)

        is_hash_approved = safe.approved_hashes(owner.address, tx_hash)
        print(f'Approved hash bytes: {tx_hash.hex()} FROM {owner.address}, status: {is_hash_approved}')

    print(f'Signature bytes (len = {len(signature_bytes)}): {signature_bytes.hex()}')

    move_tx = safe.exec_transaction(
        TO,
        VALUE,
        DATA,
        OP,
        TX_GAS,
        BASE_GAS,
        GAS_PRICE,
        GAS_TOKEN,
        REFUND,
        signature_bytes, {
            'from': ew3.wallet_address
        })

    move_tx['gas'] = move_tx.get('gas', 200000) * 2

    r = ew3.eth.send_transaction(move_tx)
    print(f'Transfer hash: {r.hex()}')

    new_balance = weth.balance_of_float(RESCUE_RECIPIENT)
    print(f'New balance: {new_balance}')
